/**
 * Copyright (C) 2010-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.metacdn.platform;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.util.Map;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpMethod;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.methods.EntityEnclosingMethod;
import org.apache.commons.httpclient.methods.StringRequestEntity;
import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.metacdn.MetaCDN;

public class MetaCDNMethod {
	static private final Logger logger = Logger.getLogger(MetaCDNMethod.class);
	
	static public final String META_CDN_URL = "http://www.metacdn.com/api";
	
	private MetaCDNAction 	   action 	   = null;
	private int                attempts    = 0;
	private String             body        = null;
	private String             contentType = null;
	private Map<String,String> headers     = null;
	private MetaCDN            provider    = null;
	
	public MetaCDNMethod(MetaCDN provider, MetaCDNAction action, Map<String,String> headers, String contentType, String body) {
		this.action = action;
		this.headers = headers;
		this.contentType = contentType;
		this.body = body;
		this.provider = provider;
	}

	@SuppressWarnings("restriction")
	String invoke(String... args) throws MetaCDNException, CloudException, InternalException {
		if (logger.isDebugEnabled()) {
			logger.debug("invoke(String...) - enter");
		}
		
		HttpMethod method;
		HttpClient client;
		int status;
		StringBuilder url = new StringBuilder();
		
		url.append(META_CDN_URL);
		url.append(action.getPrefix());		
		url.append("/");
		url.append(new String(provider.getContext().getStoragePublic()));	
		url.append(action.getType());

		if( args != null && args.length > 0 && !args[0].isEmpty()) {
			url.append("/");
			url.append(args[0]);	
		}
		
		method = action.getMethod(url.toString());		
		method.setRequestHeader( "Authorization", "Basic " + (new sun.misc.BASE64Encoder()).encode( provider.getAuthorization() ) );
		
		if (contentType != null) {
			method.setRequestHeader("Accept", contentType);
		}
		
		if( headers != null ) {
			for( Map.Entry<String, String> entry : headers.entrySet() ) {
				method.addRequestHeader(entry.getKey(), entry.getValue());
			}
		}
		if( body != null) {
			if( method instanceof EntityEnclosingMethod ) {
				try {
					((EntityEnclosingMethod)method).setRequestEntity(new StringRequestEntity(body, contentType, "utf-8"));
				} 
				catch( UnsupportedEncodingException e ) {
					logger.error(e);
					e.printStackTrace();
					throw new InternalException(e);
				}
			}
		}
		attempts++;
        client = new HttpClient();
        if( provider.getProxyHost() != null ) {
            client.getHostConfiguration().setProxy(provider.getProxyHost(), provider.getProxyPort());
        }
        
        String response = null;
        
        try {
			try {
				status =  client.executeMethod(method);
			} 
			catch( HttpException e ) {
				logger.error(e.getMessage(), e);
				throw new CloudException(e);
			} 
			catch( IOException e ) {
				logger.error(e.getMessage(), e);
				throw new InternalException(e);
			}
			if( status == HttpStatus.SC_OK || status == HttpStatus.SC_CREATED || status == HttpStatus.SC_ACCEPTED ) {
				try {
					InputStream input = method.getResponseBodyAsStream();
	
					try {
						response = parseResponse(input);
						return response;
					}
					finally {
						input.close();
					}
				} 
				catch( IOException e ) {
					logger.error(e);
					e.printStackTrace();
					throw new CloudException(e);
				}
			}
			else if( status == HttpStatus.SC_NO_CONTENT ) {
				return null;
			}
			else {
				if( status == HttpStatus.SC_SERVICE_UNAVAILABLE || status == HttpStatus.SC_INTERNAL_SERVER_ERROR ) {
					try {
						InputStream input = method.getResponseBodyAsStream();
						
						if( input != null ) {
							input.close();
						}
					} 
					catch( IOException ignore ) {
						logger.warn("IO Exception trying to close error connection: " + ignore.getMessage());
					}
					if( attempts >= 5 ) {
						String msg;
						
						if( status == HttpStatus.SC_SERVICE_UNAVAILABLE ) {
							msg = "Cloud service is currently unavailable.";
						}
						else {
							msg = "The cloud service encountered a server error while processing your request.";
						}
						logger.error(msg);
						throw new CloudException(msg);
					}
					else {
						try { Thread.sleep(5000L); }
						catch( InterruptedException e ) { }
						return invoke(args);
					}
				}
				try {
					InputStream input = method.getResponseBodyAsStream();
					String doc;
	
					try {
						doc = parseResponse(input);
					}
					finally {
						input.close();
					}
					if( doc != null ) {
						String code = null, message = null, requestId = null, type = null;
						logger.warn("Error " + doc.toString());
						//TODO identify error condition!
//						NodeList blocks = doc.getElementsByTagName("Error");
//	
//						if( blocks.getLength() > 0 ) {
//							Node error = blocks.item(0);
//							NodeList attrs;
//							
//							attrs = error.getChildNodes();
//							for( int i=0; i<attrs.getLength(); i++ ) {
//								Node attr = attrs.item(i);
//								
//								if( attr.getNodeName().equals("Code") ) {
//									code = attr.getFirstChild().getNodeValue().trim();
//								}
//								else if( attr.getNodeName().equals("Type") ) {
//									type = attr.getFirstChild().getNodeValue().trim();
//								}
//								else if( attr.getNodeName().equals("Message") ) {
//									message = attr.getFirstChild().getNodeValue().trim();
//								}
//							}
//							
//						}
//						blocks = doc.getElementsByTagName("RequestId");
//						if( blocks.getLength() > 0 ) {
//							Node id = blocks.item(0);
//							
//							requestId = id.getFirstChild().getNodeValue().trim();
//						}
						if( message == null ) {
							throw new CloudException("Unable to identify error condition: " + status + "/" + requestId + "/" + code);
						}
						throw new MetaCDNException(status, requestId, type, code, message);
					}
					throw new CloudException("Unable to parse error.");
				} 
				catch( IOException e ) {
					logger.error(e);
					e.printStackTrace();
					throw new CloudException(e);
				}			
			}
        }
		finally {
			method.releaseConnection();
			if (logger.isDebugEnabled()) {
				logger.debug("invoke(String...) - exit");
			}
		}
	}
	
	private String parseResponse(InputStream responseBodyAsStream) throws CloudException, InternalException {
		try {
			BufferedReader in = new BufferedReader(new InputStreamReader(responseBodyAsStream));
			StringBuilder sb = new StringBuilder();
			String line;
	            
			while( (line = in.readLine()) != null ) {
				sb.append(line);
			}
			in.close();
	            
			return sb.toString();
		}
		catch( IOException e ) {
			logger.error(e);
			e.printStackTrace();
			throw new CloudException(e);
		}	
	}
}
