/**
 * Copyright (C) 2010-2011 enStratusNetworks LLC
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.metacdn.util;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpMethod;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.methods.EntityEnclosingMethod;
import org.apache.commons.httpclient.methods.StringRequestEntity;
import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.metacdn.platform.MetaCDNAction;
import org.dasein.cloud.metacdn.platform.MetaCDNException;
import org.dasein.cloud.metacdn.platform.MetaCDNMethod;
import org.dasein.util.CalendarWrapper;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

/**
 * <p>Utilities for configuring MetaCDN.</p>
 * 
 * <p><b>How to run from Maven:</b><br />
 * mvn exec:java -Dexec.mainClass="org.dasein.cloud.metacdn.util.MetaCDNTestUtil" -Dexec.args="arg0 arg1 arg2"  
 * </p>
 * 
 * 
 * @author morgan
 * @author brobergj
 */

public class MetaCDNConfigUtil {
	private static final Logger logger = Logger.getLogger(MetaCDNConfigUtil.class);
	
	public static void main(String... args) {
		
		if (args == null || args.length == 0) {
			logger.fatal("Please provide arguments: cmd cmdarg1 cmdarg2 etc");
		}
		
		try {
			if (args[0].equals("create-user")) {
				createUser(args);
			} else if (args[0].equals("get-user")) {
				getUser(args);
			} else if (args[0].equals("update-user")) {
				updateUser(args);
			} else if (args[0].equals("delete-user")) {
				deleteUser(args);
			} else if (args[0].equals("create-content")) {
				createContent(args);
			} else if (args[0].equals("list-content")) {
				listContent(args);
			} else if (args[0].equals("get-content")) {
				getContent(args);
			} else if (args[0].equals("delete-content")) {
				deleteContent(args);
			} else if (args[0].equals("create-media")) {
				createMedia(args);
			} else if (args[0].equals("list-media")) {
				listMedia(args);
			} else if (args[0].equals("get-media")) {
				getMedia(args);
			} else if (args[0].equals("delete-media")) {
				deleteMedia(args);
			} else if (args[0].equals("update-media")) {
				updateMedia(args);
			} else if (args[0].equals("create-branding")) {
				createBranding(args);
			} else if (args[0].equals("list-branding")) {
				listBranding(args);
			} else if (args[0].equals("get-branding")) {
				getBranding(args);
			}
		} catch (Exception e) {
			logger.fatal(e.getMessage(), e);
		}		
	}
	
	/**
	 * args[1] should be partner username.
	 * args[2] should be partner key.
	 * args[3] should be username.
	 * args[4] should be firstname.
	 * args[5] should be lastname.
	 * args[6] should be email.
	 * args[6 + n] can be:
 	 * 
	 * <ul> 
	 *   <li>--qosPreset:One of "HIGH_TP", "LOW_RESP", "HIGH_AVAIL", "BALANCED" or "NONE"..</li>
	 *   <li>--webPropertyId:A valid Google Analytics Web Property ID in the format UA-XXXXXX-YY.</li>
	 *   <li>--domain:The origin domain you intend to serve your content from (e.g. mysite.com).</li>
	 * </ul>	 
	 * 
	 * <p><b>How to run from Maven:</b><br />
	 * mvn exec:java -Dexec.mainClass="org.dasein.cloud.metacdn.util.MetaCDNConfigUtil" -Dexec.args="create-user partnerName partnerKey username Firstname Lastname email@domain.com --qosPreset:NONE --webPropertyId:UA-XXXXXX-YY --domain:http://www.domain.com"
	 * </p> 
 	 * @param args
 	 */
	private static void createUser(String... args) throws CloudException, InternalException, MetaCDNException {
		
		JSONObject mcdnUser = new JSONObject();   	
    	try {
			mcdnUser.put("username", args[3]);
			mcdnUser.put("firstname", args[4]);
			mcdnUser.put("lastname", args[5]);
			mcdnUser.put("email", args[6]);	
			if (args.length > 7) {    		
	    		for (int i = 7; i < args.length; i++) {
	        		String arg = args[i];
	        		if (arg.contains("--qosPreset")) {
	        			// TODO: Should really check for valid preset here
	        			mcdnUser.put("qosPreset", arg.substring(12));
	        		} else if (arg.contains("--webPropertyId")) {
	        			// TODO: Should really check for valid web property id (UA-XXXXXX-YY) here
	        			mcdnUser.put("webPropertyId", arg.substring(16));
	        		} else if (arg.contains("--domain")) {
	        			// TODO: Should really check for valid domain here
	        			mcdnUser.put("domain", arg.substring(9));
	        		}
	    		}   
			}	
			logger.info(mcdnUser.toString(3));
    	} catch (JSONException e) {
    		logger.info(e.toString());
    	}
    	    	
    	String response = invoke(MetaCDNAction.CREATE_USER, getAuthorization(args), "application/json", mcdnUser.toString(), null, null);
		logger.info(response);
	}
	
	/**
 	 * args[1] should be username.
 	 * args[2] should be user key.
	 * 
	 * @param args
	 */
	private static void getUser(String... args) throws CloudException, InternalException, MetaCDNException {
		logger.info("JSON version:");
		String response = invoke(MetaCDNAction.GET_USER, getAuthorization(args), "application/json", null, args[1], null);
    	logger.info(response);
	}

	/**
	 * args[1] should be username.
	 * args[2] should be key.	
	 * args[2 + n] can be:<br />
	 * <ul>
	 *   <li>--firstname:A first name.</li> 
	 *   <li>--lastname:A last name.</li>
	 *   <li>--email:A valid email.</li>
	 *   <li>--qosPreset:One of "HIGH_TP", "LOW_RESP", "HIGH_AVAIL", "BALANCED" or "NONE"..</li>
	 *   <li>--webPropertyId:A valid Google Analytics Web Property ID in the format UA-XXXXXX-YY.</li>
	 *   <li>--domain:The origin domain you intend to serve your content from (e.g. mysite.com).</li>
	 * </ul>
	 * 
	 * <p><b>How to run from Maven:</b><br />
	 * mvn exec:java -Dexec.mainClass="org.dasein.cloud.metacdn.util.MetaCDNConfigUtil" -Dexec.args="update-user username secretKey --firstname:Joe --lastname:Bloggs --qosPreset:NONE --domain:mydomain.com.au --webPropertyId:UA-XXXXXXX-2"
	 * </p>
	 * 
	 */	
	private static void updateUser(String... args) throws CloudException, InternalException, MetaCDNException {
	
		StringBuilder sb = new StringBuilder("{");
		if (args.length > 3) {    		
    		for (int i = 3; i < args.length; i++) {
        		String arg = args[i];
        		if (arg.contains("--firstname")) {
        			sb.append("\"firstname\" : \"");
                	sb.append(arg.substring(12));
                	sb.append("\"");
        		} else if (arg.contains("--lastname")) {
        			sb.append(",\"lastname\" : \"");
                	sb.append(arg.substring(11));
                	sb.append("\"");
        		} else if (arg.contains("--email")) {
        			sb.append(",\"email\" : \"");
                	sb.append(arg.substring(8));
                	sb.append("\"");
        		} else if (arg.contains("--qosPreset")) {
        			// TODO: Should really check for valid preset here
        			sb.append(",\"qosPreset\" : \"");
                	sb.append(arg.substring(12));
                	sb.append("\"");
        		} else if (arg.contains("--webPropertyId")) {
        			// TODO: Should really check for valid web property id (UA-XXXXXX-YY) here
        			sb.append(",\"webPropertyId\" : \"");
                	sb.append(arg.substring(16));
                	sb.append("\"");
        		} else if (arg.contains("--domain")) {
        			// TODO: Should really check for valid domain here
        			sb.append(",\"domain\" : \"");
                	sb.append(arg.substring(9));
                	sb.append("\"");
        		}
    		}    		
    	}
    	
    	sb.append("}");
		
    	String response = invoke(MetaCDNAction.UPDATE_USER, getAuthorization(args), "application/json", sb.toString(), args[1], null);
		logger.info(response);				
	}	
	
	/**
	 * args[1] should be partner username.
	 * args[2] should be partner key.
	 * args[3] should be username.
	 * 
	 * @param args
	 */
	private static void deleteUser(String... args) throws CloudException, InternalException, MetaCDNException {
		String response = invoke(MetaCDNAction.DELETE_USER, getAuthorization(args), "application/json", null, args[3], null);
    	logger.info(response);
	}	
	
	/**
	 * args[1] should be username.
	 * args[2] should be user key.
	 * args[3] should be origin.
	 * args[4] should be name.
	 * 
	 * @param args
	 */
	private static void createContent(String... args) throws CloudException, InternalException, MetaCDNException {
		DateFormat df = new SimpleDateFormat("EEE, dd MMM yyyy hh:mm:ss zzz");
		
		StringBuilder sb = new StringBuilder("{\"originSource\": [{ \"originUrl\" :\"");
    	sb.append(args[3]);
    	sb.append("\", \"description\":\"");
    	sb.append(args[4]);
    	sb.append("\"}], \"locations\":[USA_WEST, USA_CENTRAL, USA_EAST, ASIA_SOUTH_EAST, ASIA_EAST, EU_WEST,EU_CENTRAL, EU_NORTH, AUS_EAST],");
    	sb.append("\"hostUntil\" : \"");
    	CalendarWrapper cal = new CalendarWrapper().getMidnight();
    	cal.calendar.add(Calendar.MONTH, 1);   //host for 1 month for now
    	sb.append(df.format(cal.getDate()));
    	sb.append("\",\"duplicates\" : true}");
		
		String response = invoke(MetaCDNAction.CREATE_CONTENT, getAuthorization(args), "application/json", sb.toString(), args[1], null);
		logger.info(response);
	}
	
	/**
	 * args[1] should be username.
	 * args[2] should be user key.
	 * 
	 * @param args
	 * @throws JSONException 
	 */
	private static void listContent(String... args) throws CloudException, InternalException, MetaCDNException, JSONException {
		String response = invoke(MetaCDNAction.LIST_CONTENT, getAuthorization(args), "application/json", null, args[1], null);
    		
		// let's try to make some JSON Objs!
		JSONArray jsonArray = new JSONArray(response);
		
		logger.info("You have " + jsonArray.length() + " object(s) configured in MetaCDN");
		
		for (int i = 0; i < jsonArray.length(); i++) {
			logger.debug((i + 1) + ". " + jsonArray.getJSONObject(i).getString("metaCdnUrl"));
		}
	}
	
	/**
	 * args[1] should be username.
	 * args[2] should be user key.
	 * args[3] should be content id.
	 * 
	 * @param args
	 */
	private static void getContent(String... args) throws CloudException, InternalException, MetaCDNException {
		logger.info("JSON version:");
		String response = invoke(MetaCDNAction.GET_CONTENT, getAuthorization(args), "application/json", null, args[1], args[3]);
    	logger.info(response);
    	
    	// TODO: There is no auth for HTTP version
    	logger.info("HTTP version:");
		response = invoke(MetaCDNAction.GET_CONTENT, null, null, null, args[1], args[3]);
    	logger.info(response);
	}
	
	/**
	 * args[1] should be username.
	 * args[2] should be user key.
	 * args[3] should be content id.
	 * 
	 * @param args
	 */
	private static void deleteContent(String... args) throws CloudException, InternalException, MetaCDNException {
		String response = invoke(MetaCDNAction.DELETE_CONTENT, getAuthorization(args), "application/json", null, args[1], args[3]);
    	logger.info(response);
	}
	
	/**
	 * args[1] should be username.
	 * args[2] should be user key.
	 * args[3] should be origin.
	 * args[4] should be aspect ratio. Either STANDARD or WIDESCREEN.
	 * args[5] should be host until. A date formatted as MM/DD/YYYY.
	 * args[6] should be locations. Any combination of USA_WEST,USA_CENTRAL,USA_EAST,ASIA_SOUTH_EAST,ASIA_EAST,EU_WEST,EU_CENTRAL,EU_NORTH,AUS_EAST.
	 * args[7] should be formats. Any combination of "mp4","flv","webm","ogg",3gp".
	 * args[8] should be quality. Any combination of "LOW","MEDIUM","HIGH".
	 * args[9] should be type. Can be Video or Audio.
	 * args[10] should be duplicates. Can be true or false.
	 * args[11] should be title.
	 * args[12] should be description.
	 * args[n+12] can be template or tags. Optional. Specified as --template:sometemplatename. Specified as --tags:"tag1","tag2","tag3"
	 * 
	 * @param args
	 * @throws ParseException 
	 */
	private static void createMedia(String... args) throws CloudException, InternalException, MetaCDNException, ParseException {
		DateFormat fromDf = new SimpleDateFormat("MM/dd/yyyy");
		DateFormat toDf = new SimpleDateFormat("EEE, dd MMM yyyy hh:mm:ss zzz");
		
		StringBuilder sb = new StringBuilder("{\"originSource\": [{ \"originUrl\" :\"");
    	sb.append(args[3]);
    	sb.append("\", \"aspectRatio\":\"");
    	sb.append(args[11]);
    	sb.append("\", \"title\":\"");
    	sb.append(args[12]);
    	sb.append("\", \"description\":\"");
    	sb.append(args[4]);
    	sb.append("\"}], \"locations\":[");
    	sb.append(args[6]);
    	sb.append("],\"hostUntil\" : \"");
    	CalendarWrapper cal = new CalendarWrapper().getMidnight();
    	cal.calendar.setTime(fromDf.parse(args[5]));
    	sb.append(toDf.format(cal.getDate()));
    	sb.append("\",\"outputFormats\" : [");
    	sb.append(args[7]);
    	sb.append("],\"quality\" : [");
    	sb.append(args[8]);
    	sb.append("],\"mediaType\" : \"");
    	sb.append(args[9]);
    	sb.append("\",\"duplicates\" : \"");
    	sb.append(args[10]);
    	sb.append("\"");
    	
    	if (args.length > 13) {    		
    		for (int i = 13; i < args.length; i++) {
        		String arg = args[i];
        		if (arg.contains("--template")) {
        			sb.append(",\"mediaBrandingTemplateName\" : \"");
                	sb.append(arg.substring(11));
                	sb.append("\"");
        		} else if (arg.contains("--tags")) {
            		sb.append(",\"tags\" : [");
            		sb.append(arg.substring(7));
                	sb.append("]");
        		}
    		}    		
    	}
    	
    	sb.append("}");
		
    	System.out.println(sb.toString());
    	
    	String response = invoke(MetaCDNAction.CREATE_MEDIA, getAuthorization(args), "application/json", sb.toString(), args[1], null);
		logger.info(response);
	}
	
	/**
	 * args[1] should be username.
	 * args[2] should be user key.
	 * 
	 * @param args
	 * @throws JSONException 
	 */
	private static void listMedia(String... args) throws CloudException, InternalException, MetaCDNException, JSONException {
		String response = invoke(MetaCDNAction.LIST_MEDIA, getAuthorization(args), "application/json", null, args[1], null);
    		
		// let's try to make some JSON Objs!
		JSONArray jsonArray = new JSONArray(response);
		
		logger.info("You have " + jsonArray.length() + " object(s) configured in MetaCDN");
		
		for (int i = 0; i < jsonArray.length(); i++) {
			logger.debug((i + 1) + ". " + jsonArray.getJSONObject(i).getString("metaCdnUrl"));
		}
	}
	
	/**
	 * args[1] should be username.
	 * args[2] should be user key.
	 * args[3] should be media id.
	 * 
	 * @param args
	 */
	private static void getMedia(String... args) throws CloudException, InternalException, MetaCDNException {
		logger.info("JSON Version:");
		String response = invoke(MetaCDNAction.GET_MEDIA, getAuthorization(args), "application/json", null, args[1], args[3]);
    	logger.info(response);
    	
    	
    	logger.info("HTML Version:");
    	response = invoke(MetaCDNAction.GET_MEDIA, getAuthorization(args), "text/html", null, args[1], args[3]);
    	logger.info(response);
	}
	
	/**
	 * args[1] should be username.
	 * args[2] should be user key.
	 * args[3] should be media id.
	 * 
	 * @param args
	 */
	private static void deleteMedia(String... args) throws CloudException, InternalException, MetaCDNException {
		String response = invoke(MetaCDNAction.DELETE_MEDIA, getAuthorization(args), "application/json", null, args[1], args[3]);
    	logger.info(response);
	}
	
	/**
	 * args[1] should be username.<br />
	 * args[2] should be user key.<br />
	 * args[3] should be media id.<br />
	 * <br />
	 * args[4 + n] can be:<br />
	 * <ul>
	 *   <li>--hostUntil:A date formatted as MM/DD/YYYY.</li>
	 *   <li>--locations:Any combination of USA_WEST,USA_CENTRAL,USA_EAST,ASIA_SOUTH_EAST,ASIA_EAST,EU_WEST,EU_CENTRAL,EU_NORTH,AUS_EAST.</li>
	 *   <li>--formats:Any combination of "mp4","flv","webm","ogg",3gp".</li>
	 *   <li>--quality:Any combination of "LOW","MEDIUM","HIGH".</li>
	 *   <li>--duplicates:Can be true or false.</li>
	 *   <li>--template:A branding template name.</li>
	 *   <li>--tags:Some tags like "tag1","tag2","tag3".</li>
	 * </ul>
	 * 
	 * @param args
	 * @throws ParseException 
	 */
	private static void updateMedia(String... args) throws CloudException, InternalException, MetaCDNException, ParseException {
		DateFormat fromDf = new SimpleDateFormat("MM/dd/yyyy");
		DateFormat toDf = new SimpleDateFormat("EEE, dd MMM yyyy hh:mm:ss zzz");
		
		StringBuilder sb = new StringBuilder("{");
		if (args.length > 3) {    		
    		for (int i = 4; i < args.length; i++) {
        		String arg = args[i];
				if (arg.contains("--hostUntil")) {
					sb.append("\"hostUntil\" : \"");
					CalendarWrapper cal = new CalendarWrapper().getMidnight();
					cal.calendar.setTime(fromDf.parse(arg.substring(12)));
					sb.append(toDf.format(cal.getDate()));
					sb.append("\"");
				} else if (arg.contains("--locations")) {
					if (sb.length() > 1) {
						sb.append(",");
					}
					sb.append("\"locations\" : [");
					sb.append(arg.substring(12));
					sb.append("]");
				} else if (arg.contains("--formats")) {
					if (sb.length() > 1) {
						sb.append(",");
					}
					sb.append("\"formats\" : [");
					sb.append(arg.substring(10));
					sb.append("]");
				} else if (arg.contains("--quality")) {
					if (sb.length() > 1) {
						sb.append(",");
					}
					sb.append("\"quality\" : [");
					sb.append(arg.substring(10));
					sb.append("]");
				} else if (arg.contains("--duplicates")) {
					if (sb.length() > 1) {
						sb.append(",");
					}
					sb.append("\"duplicates\" : \"");
					sb.append(arg.substring(13));
					sb.append("\"");
				} else if (arg.contains("--template")) {
					if (sb.length() > 1) {
						sb.append(",");
					}
        			sb.append("\"mediaBrandingTemplateName\" : \"");
                	sb.append(arg.substring(11));
                	sb.append("\"");
        		} else if (arg.contains("--tags")) {
        			if (sb.length() > 1) {
						sb.append(",");
					}
            		sb.append("\"tags\" : [");
            		sb.append(arg.substring(7));
                	sb.append("]");
        		} else if (arg.contains("--title")) {
        			if (sb.length() > 1) {
						sb.append(",");
					}
            		sb.append("\"title\" : \"");
            		sb.append(arg.substring(8));
                	sb.append("\"");
        		}  else if (arg.contains("--description")) {
        			if (sb.length() > 1) {
						sb.append(",");
					}
            		sb.append("\"description\" : \"");
            		sb.append(arg.substring(14));
                	sb.append("\"");
        		}
    		}    		
    	}
    	
    	sb.append("}");
    	
    	System.out.println(sb.toString());
		
    	String response = invoke(MetaCDNAction.UPDATE_MEDIA, getAuthorization(args), "application/json", sb.toString(), args[1], args[3]);
		logger.info(response);
	}
	
	/**
	 * args[1] should be username.
	 * args[2] should be user key.
	 * args[3] path to a branding template JSON string.  Let's not bother trying to build a string and customize an example from MetaCDN. 
	 * 
	 * @param args
	 * @throws ParseException 
	 * @throws IOException 
	 */
	private static void createBranding(String... args) throws CloudException, InternalException, MetaCDNException, ParseException, IOException {
		StringBuilder sb = new StringBuilder();
    	
		BufferedReader in = null;
		
		try {
			in = new BufferedReader(new FileReader(args[3]));
			String line;
			
			while ((line = in.readLine())  != null) {
				sb.append(line);
			}
		} finally {
			in.close();
		}
		
    	String response = invoke(MetaCDNAction.CREATE_BRANDING, getAuthorization(args), "application/json", sb.toString(), args[1], null);
		logger.info(response);
	}
	
	/**
	 * args[1] should be username.
	 * args[2] should be user key.
	 * 
	 * @param args
	 * @throws JSONException 
	 */
	private static void listBranding(String... args) throws CloudException, InternalException, MetaCDNException, JSONException {
		String response = invoke(MetaCDNAction.LIST_BRANDING, getAuthorization(args), "application/json", null, args[1], null);
    		
		// let's try to make some JSON Objs!
		JSONArray jsonArray = new JSONArray(response);
		
		logger.info("You have " + jsonArray.length() + " object(s) configured in MetaCDN");
		
		for (int i = 0; i < jsonArray.length(); i++) {
			logger.debug((i + 1) + ". " + jsonArray.getJSONObject(i).getString("metaCdnUrl"));
		}
	}
	
	/**
	 * args[1] should be username.
	 * args[2] should be user key.
	 * args[3] should be branding template name.
	 * 
	 * @param args
	 */
	private static void getBranding(String... args) throws CloudException, InternalException, MetaCDNException {
		logger.info("JSON Version:");
		String response = invoke(MetaCDNAction.GET_BRANDING, getAuthorization(args), "application/json", null, args[1], args[3]);
    	logger.info(response);
    	
    	
    	logger.info("HTML Version:");
    	response = invoke(MetaCDNAction.GET_BRANDING, getAuthorization(args), "text/html", null, args[1], args[3]);
    	logger.info(response);
	}
	
	private static byte[] getAuthorization(String... args) {
    	StringBuilder sb = new StringBuilder(args[1]);
    	sb.append(":");
    	sb.append(args[2]);
    	return sb.toString().getBytes();
    }
	
	private static String invoke(MetaCDNAction action, byte[] authorization, String contentType, String body, String user, String id) throws CloudException, InternalException, MetaCDNException {
		StringBuilder url = new StringBuilder();
		
		HttpMethod method;
		HttpClient client;
		int status;
		
		url.append(MetaCDNMethod.META_CDN_URL);
		
		url.append(action.getPrefix());
		
		if (user != null) {
			url.append("/");
			url.append(user);
		}
		
		url.append(action.getType());
		
		if( id != null ) {
			url.append("/");
			url.append(id);	
		}
		
		method = action.getMethod(url.toString());		
		method.setRequestHeader( "Authorization", "Basic " + (new sun.misc.BASE64Encoder()).encode( authorization ) );
		if (contentType != null) {
			method.setRequestHeader("Accept", contentType);
		}
		
		if( body != null) {
			if( method instanceof EntityEnclosingMethod ) {
				try {
					((EntityEnclosingMethod)method).setRequestEntity(new StringRequestEntity(body, contentType, "utf-8"));
				} 
				catch( UnsupportedEncodingException e ) {
					throw new InternalException(e);
				}
			}
		}
		client = new HttpClient();
        String response = null;
        
        try {
			try {
				status =  client.executeMethod(method);
			} 
			catch( HttpException e ) {
				throw new CloudException(e);
			} 
			catch( IOException e ) {
				throw new InternalException(e);
			}
			if( status == HttpStatus.SC_OK || status == HttpStatus.SC_CREATED || status == HttpStatus.SC_ACCEPTED ) {
				try {
					InputStream input = method.getResponseBodyAsStream();
	
					try {
						response = parseResponse(input);
						return response;
					}
					finally {
						input.close();
					}
				} 
				catch( IOException e ) {
					throw new CloudException(e);
				}
			}
			else if( status == HttpStatus.SC_NO_CONTENT ) {
				return null;
			}
			else {
				if( status == HttpStatus.SC_SERVICE_UNAVAILABLE || status == HttpStatus.SC_INTERNAL_SERVER_ERROR ) {
					try {
						InputStream input = method.getResponseBodyAsStream();
						
						if( input != null ) {
							input.close();
						}
					} 
					catch( IOException ignore ) {
						logger.error("IO Exception trying to close error connection: " + ignore.getMessage());
					}
					
				}
				try {
					InputStream input = method.getResponseBodyAsStream();
					String doc;
	
					try {
						doc = parseResponse(input);
						logger.info(doc);
					}
					finally {
						input.close();
					}
					if( doc != null ) {
						String code = null, message = null, requestId = null, type = null;
						//TODO identify error condition!
//						NodeList blocks = doc.getElementsByTagName("Error");
//	
//						if( blocks.getLength() > 0 ) {
//							Node error = blocks.item(0);
//							NodeList attrs;
//							
//							attrs = error.getChildNodes();
//							for( int i=0; i<attrs.getLength(); i++ ) {
//								Node attr = attrs.item(i);
//								
//								if( attr.getNodeName().equals("Code") ) {
//									code = attr.getFirstChild().getNodeValue().trim();
//								}
//								else if( attr.getNodeName().equals("Type") ) {
//									type = attr.getFirstChild().getNodeValue().trim();
//								}
//								else if( attr.getNodeName().equals("Message") ) {
//									message = attr.getFirstChild().getNodeValue().trim();
//								}
//							}
//							
//						}
//						blocks = doc.getElementsByTagName("RequestId");
//						if( blocks.getLength() > 0 ) {
//							Node id = blocks.item(0);
//							
//							requestId = id.getFirstChild().getNodeValue().trim();
//						}
						if( message == null ) {
							throw new CloudException("Unable to identify error condition: " + status + "/" + requestId + "/" + code);
						}
						throw new MetaCDNException(status, requestId, type, code, message);
					}
					throw new CloudException("Unable to parse error.");
				} 
				catch( IOException e ) {
					throw new CloudException(e);
				}			
			}
        }
		finally {
			method.releaseConnection();			
		}
	}
	
	private static String parseResponse(InputStream responseBodyAsStream) throws CloudException, InternalException {
		try {
			BufferedReader in = new BufferedReader(new InputStreamReader(responseBodyAsStream));
			StringBuilder sb = new StringBuilder();
			String line;
	            
			while( (line = in.readLine()) != null ) {
				sb.append(line);
			}
			in.close();
	            
			return sb.toString();
		}
		catch( IOException e ) {
			e.printStackTrace();
			throw new CloudException(e);
		}	
	}
}